import logging
import os
from datetime import datetime, timezone
from warnings import filterwarnings

import h5netcdf
import numpy as np
import xarray as xr
from cftime import date2num, num2date

from .utils import merge_arrays


def xarray2netcdf(xArray, fdir, group="original", timedim="datetime",
                  archive_starttime=datetime(2000, 1, 1), resolution=None,
                  mode='a'):
    """
    Store an xarray dataset as an HDF5 file.

    Parameters
    ----------
    xArray : xarray.Dataset
        Data to store.
    fdir : str
        Directory to store data under.
    group : str
        Hdf5 group name.
    timedim : str
        Name of time dimension.
    archive_starttime : datetime
        Start time of archive. If the start time of the data is before this
        time, the data start time is used.
    resolution : float
        Time resolution of the data in hours. If None, the resolution is
        determined from the data.
    """
    filterwarnings(action='ignore', category=DeprecationWarning,
                   message='`np.bool` is a deprecated alias')

    data_starttime = xArray[timedim].values[0].astype(
        'datetime64[us]').astype(datetime)
    starttime = min(data_starttime, archive_starttime)
    now = datetime.now(tz=timezone.utc)
    if resolution is None:
        resolution = (np.diff(xArray[timedim])/np.timedelta64(1, 'h'))[0]

    for featureName in list(xArray.data_vars.keys()):
        h5file = os.path.join(fdir, featureName + '.nc')
        _mode = 'w'
        if os.path.isfile(h5file) and mode == 'a':
            if archive_starttime > data_starttime:
                xds_existing = xr.open_dataset(
                    h5file, group=group, engine='h5netcdf')
                xda_new = merge_arrays(
                    xds_existing[featureName], xArray[featureName],
                    resolution=resolution)
                xds_existing.close()
                xda_new.to_netcdf(h5file, group=group,
                                  mode='w', engine='h5netcdf')
                continue
            _mode = 'a'

        with h5netcdf.File(h5file, _mode) as h5f:
            try:
                rootGrp = _create_root_group(group, featureName,
                                             h5f, xArray, starttime, timedim)
            except ValueError:  # group already exists, append
                rootGrp = h5f[group]

            try:
                metaGrp = _create_meta_group(h5f, resolution)
            except ValueError:  # group already exists, append
                metaGrp = h5f['meta']

            # determine indices
            new_time = date2num(xArray[timedim].values.astype('datetime64[us]').astype(datetime),
                                units=rootGrp[timedim].attrs['units'],
                                calendar=rootGrp[timedim].attrs['calendar'])
            t0 = date2num(starttime,
                          units=rootGrp[timedim].attrs['units'],
                          calendar=rootGrp[timedim].attrs['calendar'])

            indices = np.rint((new_time - t0)/resolution).astype(int)
            if not np.all(indices >= 0):
                raise ValueError("Data starts before the archive start time")
            times = rootGrp[timedim]
            newsize = indices[-1] + 1
            if newsize > times.shape[0]:
                rootGrp.resize_dimension(timedim, newsize)
            times[:] = t0 + np.arange(times.shape[0]) * resolution
            data = rootGrp[featureName]
            if len(data.shape) > 1:
                data[:, indices] = xArray[featureName].values
            else:
                data[indices] = xArray[featureName].values
            now_time = date2num(now, units=metaGrp['update_log'].attrs['units'],
                                calendar=metaGrp['update_log'].attrs['calendar'])
            ulog = metaGrp['update_log']
            ldata = metaGrp['last_datapoint']
            metaGrp.resize_dimension('update', ulog.shape[0] + 1)
            ulog[-1] = now_time
            metaGrp.resize_dimension('endtime', ldata.shape[0] + 1)
            ldata[-1] = times[-1]
            old_resolution = metaGrp['resolution'][()]
            if old_resolution != resolution:
                raise ValueError(f"Resolution mismatch for {featureName}: "
                                 f"{old_resolution} != {resolution}")


def _create_root_group(defaultGroupName, featureName, h5f, xArray, starttime, timedim):
    rootGrp = h5f.create_group(defaultGroupName)
    rootGrp.dimensions[timedim] = None
    coordinates = rootGrp.create_variable(timedim, (timedim,), float)
    coordinates.attrs['units'] = 'hours since 1970-01-01 00:00:00.0'
    coordinates.attrs['calendar'] = 'gregorian'
    rootGrp.attrs['archive_starttime'] = str(starttime)
    for label, size in xArray.sizes.items():
        if not np.issubdtype(xArray[label].dtype, np.datetime64):
            rootGrp.dimensions[label] = size
            coordinates = rootGrp.create_variable(label, (label,), float)
            coordinates[:] = xArray[label].values
    # Note: xArray.dims returns a dictionary of dimensions that are not necesarily
    # in the right order; xArray[featureName].dims returns a tuple with dimension
    # names in the correct order
    rootGrp.create_variable(featureName, tuple(
        xArray[featureName].dims), dtype=float, fillvalue=0.)
    _set_attributes(featureName, rootGrp, xArray)
    return rootGrp


def _set_attributes(featureName, rootGrp, xArray):
    """
    Set attributes for the root group. Attributes are assumed to not change
    over time. If they do, they should be stored in the 'meta' group.
    """
    for key, value in xArray.attrs.items():
        rootGrp.attrs[key] = value
    rootGrp.attrs['feature'] = featureName


def _create_meta_group(h5f, resolution):
    """
    Create meta group to track processing history.
    """
    metaGrp = h5f.create_group('meta')
    metaGrp.dimensions['update'] = None
    ulog = metaGrp.create_variable('update_log', ('update',), float)
    ulog.attrs['units'] = 'hours since 1970-01-01 00:00:00.0'
    ulog.attrs['calendar'] = 'gregorian'
    metaGrp.dimensions['endtime'] = None
    ldata = metaGrp.create_variable('last_datapoint', ('endtime',), float)
    ldata.attrs['units'] = 'hours since 1970-01-01 00:00:00.0'
    ldata.attrs['calendar'] = 'gregorian'
    res = metaGrp.create_variable("resolution", (), dtype=float)
    res[()] = resolution
    return metaGrp



import numpy as np

from tonik.utils import extract_consecutive_integers


def test_extract_consecutive_integers():
    nums = [1, 2, 3, 5, 6, 7, 8, 10]
    assert extract_consecutive_integers(
        nums) == [[1, 2, 3], [5, 6, 7, 8], [10]]
    assert extract_consecutive_integers([1]) == [[1]]
    assert extract_consecutive_integers(np.array([1, 2, 4])) == [[1, 2], [4]]



from datetime import datetime, timezone
import logging
import os

import numpy as np
import pandas as pd
import xarray as xr
try:
    from zarr.errors import PathNotFoundError
except ImportError:
    class PathNotFoundError(Exception):
        pass

from .utils import merge_arrays, fill_time_gaps, get_dt

logger = logging.getLogger(__name__)


def get_chunks(xda: xr.DataArray, chunks: int = 1,
               timedim: str = 'datetime') -> dict:
    """
    Determine the chunk size for the datetime dimension. Other dimensions are assumed to be
    small enough to not require chunking.

    Parameters
    ----------
    coords : xr.core.coordinates.DatasetCoordinates
        Coordinates of the dataset.
    chunks : int, optional
        Number of chunks in days to divide the datetime dimension into, by default 1.
    """
    if timedim not in xda.coords:
        raise ValueError(
            f"Datetime coordinate {timedim} not found in dataset coordinates.")
    dt = get_dt(xda.coords[timedim])
    chunklength = int(pd.Timedelta('%dD' % chunks) / dt)
    return chunklength


def fill_time_gaps_between_datasets(xds_existing: xr.DataArray, xds_new: xr.DataArray, mode: str,
                                    timedim: str = 'datetime') -> xr.DataArray:
    """
    Fill gaps between existing and new datasets.

    Parameters
    ----------
    xds_existing : xr.Dataset
        Existing dataset on disk
    xds_new : xr.Dataset
        New dataset to append
    timedim : str
        Name of the time dimension

    Returns
    -------
    xr.Dataset
        Combined dataset with gaps filled
    """
    if mode not in ['a', 'p']:
        raise ValueError(
            'Mode has to be either "a" for append or "p" for prepend')

    # get the sample interval
    dt = get_dt(xds_new.coords[timedim])

    existing_endpoint = xds_existing[timedim].values
    # Get time ranges
    if mode == 'a':
        gap_start = existing_endpoint + dt
        gap_end = xds_new[timedim].values[0] - dt
    elif mode == 'p':
        gap_end = existing_endpoint - dt
        gap_start = xds_new[timedim].values[-1] + dt

    if gap_start <= gap_end:
        gap_times = pd.date_range(start=gap_start, end=gap_end, freq=dt)

        # Create NaN array with same shape as variable but for gap times
        gap_shape = (len(gap_times),) + \
            xds_new.shape[1:]  # Skip time dimension
        gap_values = np.full(gap_shape, np.nan)

        # Create coordinates for gap dataset
        gap_coords = {timedim: gap_times}
        for coord_name, coord in xds_new.coords.items():
            if coord_name != timedim:
                gap_coords[coord_name] = coord

        gap_data = xr.DataArray(
            gap_values,
            coords=gap_coords,
            dims=xds_new.dims,
            name=xds_new.name
        )

        # Combine: existing + gap + new
        if mode == 'a':
            combined = xr.concat([gap_data, xds_new], dim=timedim)
        elif mode == 'p':
            combined = xr.concat([xds_new, gap_data], dim=timedim)
        return combined
    else:
        return xds_new


def _build_append_payload_full_chunks(payload: xr.DataArray, mode: str,
                                      chunklen: int, timedim: str = "datetime") -> xr.DataArray:
    """
    Construct the sequence to append so that the final total length is a multiple of `chunklen`

    Parameters
    ----------
    payload : xr.DataArray
        DataArray to append
    mode : str
        'a' for append, 'p' for prepend
    chunklen : int
        Chunk length in number of time steps
    timedim : str
        Name of the time dimension

    Returns
    -------
    xr.DataArray
        Padded DataArray with length a multiple of chunklen
    """
    if mode not in ['a', 'p']:
        raise ValueError(
            'Mode has to be either "a" for append or "p" for prepend')

    # pad the tail so that payload_len % chunklen == 0
    pay_len = payload.sizes[timedim]
    need = -pay_len % chunklen  # 0..chunklen-1

    if need > 0:
        dt = get_dt(payload.coords[timedim])
        if mode == 'a':
            start = payload[timedim].values[-1] + dt
        elif mode == 'p':
            start = payload[timedim].values[0] - (need+1)*dt
        pad_times = pd.date_range(start=start, periods=need, freq=dt)
        pad_shape = []
        for i, d in enumerate(payload.dims):
            if d == timedim:
                pad_shape.append(need)
            else:
                pad_shape.append(payload.shape[i])
        pad_vals = np.full(pad_shape, np.nan)
        pad_coords = {timedim: pad_times}
        for c in payload.coords:
            if c != timedim:
                pad_coords[c] = payload.coords[c]
        pad_da = xr.DataArray(pad_vals, coords=pad_coords,
                              dims=payload.dims, name=payload.name, attrs=payload.attrs)
        if mode == 'a':
            payload = xr.concat([payload, pad_da], dim=timedim)
        elif mode == 'p':
            payload = xr.concat([pad_da, payload], dim=timedim)
        payload = payload.chunk({timedim: chunklen})
    return payload


def _update_meta_data(fout: str,
                      last_datapoint: np.datetime64,
                      resolution: float | None = None,
                      meta_group: str = "meta") -> None:
    """
    Append current update time (and last_datapoint) to meta group.

    Parameters
    ----------
    fout : str
        Base zarr store path (per-variable .zarr directory).
    last_datapoint : np.datetime64
        Latest data time in the feature.
    resolution : float | None
        Optional time resolution (hours) to store once.
    meta_group : str
        Group name for metadata.
    """

    now = np.datetime64(datetime.now(tz=timezone.utc), 'ns')
    new_update = xr.DataArray([now],
                              coords={'update': [now]},
                              dims=['update'],
                              name='update_log')
    new_last = xr.DataArray([last_datapoint],
                            coords={'endtime': [now]},
                            dims=['endtime'],
                            name='last_datapoint')

    try:
        meta = xr.open_zarr(fout, group=meta_group, chunks=None)
        # Existing vars -> concatenate
        update_old = meta.get('update_log')
        last_old = meta.get('last_datapoint')
        res_da_old = meta.get('resolution').values[()]
        new_update = xr.concat([update_old, new_update], dim='update')
        new_last = xr.concat([last_old, new_last], dim='endtime')
        if resolution != res_da_old:
            raise ValueError(f"Resolution mismatch for {fout}: "
                             f"{res_da_old} != {resolution}")
        res_da = xr.DataArray(resolution, name='resolution')
    except Exception:
        # First creation
        res_da = xr.DataArray(
            resolution, name='resolution') if resolution is not None else None

    vars = {'update_log': new_update, 'last_datapoint': new_last}
    if res_da is not None:
        vars['resolution'] = res_da
    xr.Dataset(vars).to_zarr(fout, group=meta_group, mode='w')


def xarray2zarr(xds: xr.Dataset, path: str, mode: str = 'a', group='original',
                chunks: int = 10, timedim: str = 'datetime') -> None:
    """
    Write xarray dataset to zarr files.

    Parameters
    ----------
    xds : xr.Dataset
        Dataset to write.
    path : str
        Path to write the dataset.
    mode : str, optional
        Write mode, by default 'a'.
    group : str, optional
        Group name, by default 'original'
    chunks : int, optional
        Chunk size as the number of days.
    timedim : str
        Name of the time dimension, by default 'datetime'
    fill_gaps : bool, optional
        Whether to fill time gaps with NaN before writing, by default False

    Returns
    -------
    None
    """

    if timedim not in xds.dims:
        raise ValueError(f"{timedim} dimension not found in Dataset.")

    # Fill gaps
    xds = xds.drop_duplicates(timedim, keep='last')
    xds = fill_time_gaps(xds, timedim=timedim)

    for feature in xds.data_vars.keys():
        fout = os.path.join(path, feature + '.zarr')
        # nchunks = get_chunks(xds[feature], chunks)
        nchunks = chunks
        last_dp = xds[feature][timedim].values[-1]
        _update_meta_data(fout, last_dp, resolution=float(
            get_dt(xds[timedim]) / pd.Timedelta(1, 'h')))
        try:
            xds_existing = xr.open_zarr(fout, group=group)
            has_store = True
        except (PathNotFoundError, FileNotFoundError, KeyError):
            has_store = False

        if not has_store:
            xda_new = _build_append_payload_full_chunks(
                xds[feature], 'a', nchunks)
            xda_new.to_zarr(fout, group=group, mode='w',
                            write_empty_chunks=True)
            continue

        if xds_existing[timedim][0] > xds[timedim][-1]:
            # prepend
            xda_new = fill_time_gaps_between_datasets(xds_existing[feature].isel({timedim: 0}),
                                                      xds[feature], mode='p')
            xda_new = _build_append_payload_full_chunks(
                xda_new, 'p', nchunks)
            combined = xda_new.combine_first(xds_existing[feature]).compute()
            combined.chunk({timedim: nchunks}).to_zarr(fout, group=group, mode='w',
                                                       write_empty_chunks=True)

        elif xds_existing[timedim][-1] < xds[timedim][0]:
            # append
            xda_new = fill_time_gaps_between_datasets(xds_existing[feature].isel({timedim: -1}),
                                                      xds[feature], mode='a')
            xda_new = _build_append_payload_full_chunks(
                xda_new, 'a', nchunks)
            xda_new.to_zarr(fout, group=group, mode='a',
                            append_dim=timedim)

        elif xds_existing[timedim][0] > xds[timedim][0] and xds_existing[timedim][-1] < xds[timedim][-1]:
            # existing datetimes are contained in new array
            xda_new = _build_append_payload_full_chunks(
                xds[feature], 'a', nchunks)
            xda_new.to_zarr(fout, group=group, mode='w',
                            write_empty_chunks=True)

        else:
            overlap = xds_existing[timedim].where(
                xds_existing[timedim] == xds[timedim])
            xds[feature].loc[{timedim: overlap}].to_zarr(
                fout, group=group, mode='r+', region='auto')
            remainder = xds[feature].drop_sel({timedim: overlap})
            if remainder.sizes[timedim] > 0:
                mode = 'a'
                if remainder[timedim][-1] < xds_existing[timedim][0]:
                    mode = 'p'
                xda_new = fill_time_gaps_between_datasets(xds_existing[feature].isel({timedim: 0}),
                                                          xds[feature], mode=mode)
                xda_new = _build_append_payload_full_chunks(
                    xda_new, mode, nchunks)
                xda_new.to_zarr(fout, group=group, mode='a',
                                append_dim=timedim)



"""
Post labels as Grafana annotations and retrieve them via the Grafana API.
Labels are assumed to be in the following format:

[
    {"time": "2023-01-01T00:00:00Z",
     "title": "Label 1",
     "description": "Some description",
     "tags": ["tag1", "tag2"],
     "id": 12345
     },
    {"time": "2023-02-01T00:00:00Z",
     "timeEnd": "2023-02-01T01:00:00Z",
     "title": "Label 2", description:
     "description": "Another description",
     "tags": ["tag1", "tag2"],
     "id": 12346
     },
     ...
]

Note: The `timeEnd` field is optional. If it is provided, the grafana annotation will 
be a range annotation, otherwise it will be a point annotation. The label id is used to
check if an annotation was already added to Grafana. If the label id is not present in the
annotations, a new annotation will be created. If the label id is present, the annotation
will be skipped to avoid duplicates.
"""

import numpy as np
import requests
import uuid
from typing import List, Dict, Any, Optional


def get_annotations(baseurl: str, tags: Optional[List[str]] = None) -> Dict[str, Dict[str, Any]]:
    """
    Get existing annotations from the Grafana API.

    Parameters
    ----------
    baseurl : str
        The base URL of the Grafana API. This is typically in the format
        "http://<grafana-user>:<user-pwd>@<grafana-host>:<port>".
    tags : list of str, optional
        A list of tags to filter the annotations. If None, all annotations are retrieved.
    Returns
    -------
    annotations : dict
        A dictionary containing the annotations, where the keys are the label IDs
        and the values are dictionaries with the following keys:
            - 'text': The text of the annotation.
            - 'time': The start time of the annotation in milliseconds since epoch.
            - 'timeEnd': (optional) The end time of the annotation in milliseconds since epoch.
            - 'tags': A list of tags associated with the annotation.
            - 'id': The ID of the annotation.
    """
    url = baseurl + "/api/annotations"
    params = {}
    if tags is not None:
        params = {"tags": tags}
    rval = requests.get(url, params=params)
    if rval.status_code != 200:
        raise RuntimeError(
            f"Failed to retrieve annotations: {rval.status_code} {rval.text}")
    annotations = {}
    for atn in rval.json():
        # Extract label ID from text using regex
        try:
            label_id = atn["text"].split("Id: ")[-1].strip()
        except IndexError:
            label_id = str(uuid.uuid4())
        annotations[label_id] = {"text": atn["text"],
                                 "time": atn["time"],
                                 "timeEnd": atn.get("timeEnd", None),
                                 "tags": atn.get("tags", []),
                                 "id": atn["id"]}
    return annotations


def post_annotations(baseurl: str, labels: List[Dict[str, Any]]) -> None:
    """
    Post annotations from a list of labels if they do not already exist.

    Parameters
    ----------
    baseurl : str
        The base URL of the Grafana API. This is typically in the format
        "http://<grafana-user>:<user-pwd>@<grafana-host>:<port>".
    labels : list of dict
        A list of dictionaries containing label information. Each dictionary
        should have the following keys:
            - 'time': The start time of the annotation in ISO format.
            - 'timeEnd': (optional) The end time of the annotation in ISO format.
            - 'title': The title of the annotation.
            - 'description': The description of the annotation.
            - 'tags': A list of tags associated with the annotation.
            - 'id': A unique identifier for the label, used to check for duplicates.
    """
    url = baseurl + "/api/annotations"
    header = {"Content-type": "application/json"}
    for label in labels:
        existing_labels = get_annotations(baseurl, tags=label['tags'])
        if str(label['id']) in existing_labels:
            print(
                f"Label with ID {label['id']} already exists. Skipping post.")
            continue
        starttime = np.datetime64(label['time']).astype(
            'datetime64[ms]').astype(int)
        try:
            endtime = np.datetime64(label['timeEnd']).astype(
                'datetime64[ms]').astype(int)
            endtime = int(endtime)
        except KeyError:
            endtime = None
        text = f"{label['title']}\n{label['description']}\nId: {label['id']}"
        new_annotation = {
            "time": int(starttime),
            "timeEnd": endtime,
            "text": text,
            "tags": label['tags']
        }
        rval = requests.post(url, headers=header, json=new_annotation)
        if rval.status_code != 200:
            raise RuntimeError(
                f"Failed to post annotation: {rval.status_code} {rval.text}")
        else:
            print("Annotation posted successfully.")


def main(argv=None):
    import argparse
    import json
    parser = argparse.ArgumentParser(
        description=__doc__)
    parser.add_argument('url', type=str,
                        help='Grafana API URL for annotations.')
    parser.add_argument('--labels', type=str, default=None,
                        help='Path to JSON file containing labels to post.')
    parser.add_argument('--get-annotations', action='store_true',
                        help='Get existing annotations from Grafana API.')
    args = parser.parse_args(argv)
    if args.labels is not None:
        with open(args.labels, 'r') as f:
            labels = json.load(f)
        post_annotations(args.url, labels)
    elif args.get_annotations:
        annotations = get_annotations(args.url)
        print(json.dumps(annotations, indent=2))


if __name__ == "__main__":
    main()
